## First all the Perlegen functions
read.Perlegen.phased <- function(chromosome,path="/data/Perlegen/")
{
  ## read from a preprocessed file as class sima
  filename <- paste(path,"chromosome",chromosome,".haplotypes",sep="")
  a <- readsima(filename);
  levels(a$locations) <- c("AA","CEU","HCB")
  class(a) <- c("Perlegen.haplotypes","simapop","sima");
  a;
}
## extract
read.Perlegen.info <- function(chromosome,path="/data/Perlegen/")
{
  ## read the information file
  if (chromosome<10) chromosome <- paste("0",chromosome,sep="")
  filename <- paste(path,"s_chr",chromosome,".dat",sep="")
  d <- read.table(filename,header=T);
  class(d) <- c("Perlegen.info","data.frame")
  d;
}
extract.Perlegen.haplotypes <- function(x,start,finish)
  {
    u <- x$positions>=start*10^6&x$positions<=finish*10^6
    a <- list(d=x$d[,u],positions=x$positions[u],locations=x$locations,SNP=x$SNP[u])
    class(a) <-  c("Perlegen.haplotypes","simapop","sima");
    a;
  }
##read.Perlegen.genotypes <- function(chromosome,path="/data/Perlegen/")
##{
##  if (chromosome<10) chromosome <- paste("0",chromosome,sep="")
##  filename <- paste(path,"h_chr",chromosome,".dat",sep="")
###  d <- read.table(positions <- lapply(legendfilename,function(x) read.table(gzfile(x),header=T));
##  length <- unlist(lapply(positions,nrow))
##  pos <- lapply(positions,function(x) x[,1])
##  datafilename <- paste("/data/hapmap/phasing/full/genotypes_chr",chromosome,"_",pops,".phased.gz",sep="")
##  names(datafilename) <- pops;
##  d <- lapply(datafilename,function(x) scan(gzfile(x)))
##  for (i in 1:length(d)) 
##    d[[i]] <- matrix(d[[i]],ncol=length[i],byrow=T)

##  res <- list(d=d,positions=pos,n=length)
##  class(res) <- c("hapmap.haplotypes","haplotypes")
##  res;
##}
## functions to read the standard haplotype and genotype files



read.hapmap.phasing.old <- function(chromosome,pops=c("CEU","HCB")
                                ,path="/data/hapmap/phasing/full/",collapse=FALSE)
{
  cat("Reading haplotypes from chromosome",chromosome,"for populations",paste(pops),"\n")

  ## first read in the legends
  legendfilename <- paste(path,"genotypes_chr",chromosome,"_",pops,"_legend.txt.gz",sep="")

  names(legendfilename) <- pops;
  positions <- lapply(legendfilename,function(x)
                      read.table(gzfile(x),header=T))
  length <- unlist(lapply(positions,nrow))
  pos <- lapply(positions,function(x) x[,1])
  SNP <- lapply(positions,function(x)
                paste(levels(x[,2])[x[,2]],"/",levels(x[,3])[x[,3]],sep=""))
  ## read in the monomorphic sites
  monofilename <- paste("/data/hapmap/phasing/full/genotypes_chr",chromosome,"_",pops,"_monomorphic.txt.gz",sep="")
 
  names(monofilename) <- pops;
  monomorphic <- lapply(monofilename,function(x) read.table(gzfile(x)))
  monopos <-  lapply(monomorphic,function(x) x[,1])
  monoSNP <-  lapply(monomorphic,function(x) paste(x[,2]))
  
   ## read in the polymorphic data
  datafilename <- paste("/data/hapmap/phasing/full/genotypes_chr",chromosome,"_",pops,".phased.gz",sep="")
  names(datafilename) <- pops;
  d <- lapply(datafilename,function(x) scan(gzfile(x)))
  for (i in 1:length(d)) 
    d[[i]] <- matrix(d[[i]],ncol=length[i],byrow=T)
  
  if (length(pops)==1) {
    ## do not include the monomorphic data
    res <- list(d=d[[1]],positions=pos[[1]],data.source="HAPMAP"
                , label=paste("HAPMAP data from chromosome",chromosome,"from population",pops))
    class(res) <- "GenomicHaplotype"
    return(res)
  }
  
  if (collapse) {
    return(extract.hapmap.haplotypes(
                                     list(d=d,positions=pos,n=length,SNP=SNP,monomorphic)
                                     ,start=0,finish=1000)
           )
  } else {
    res <- list(d=d,positions=pos,n=length,SNP=SNP,monopos=monopos,monoSNP=monoSNP)
    class(res) <- c("hapmap.haplotypes","haplotypes")
    return(res);
  }
}

read.hapmap.phasing <- function(chromosome,pops=c("CEU","HCB")
                                ,path="/data/hapmap/phasing/full/",collapse=FALSE)
{
  read.phasing <- function(pop) {
    stem <-  paste(path,"genotypes_chr",chromosome,"_",pop,sep="")
   # conn <- gzfile(paste(stem,"_legend.txt.gz",sep=""))
   # print(showConnections(all=T))
    positions <-  read.table(gzfile(paste(stem,"_legend.txt.gz",sep="")),header=T)
  #   print(showConnections(all=T))
 #   close(conn)
    length <- nrow(positions)
    mono <-  read.table(gzfile(paste(stem,"_monomorphic.txt.gz",sep="")))
   # print(showConnections(all=T))

    d <-  matrix(scan( conn <- gzfile(paste(stem,".phased.gz",sep=""))),ncol=length,byrow=T)
    close(conn)
    r <- list(d=d,positions=positions[,1],SNP=paste(positions[,2],positions[,3],sep="/"),monopos=mono[,1],monoSNP=paste(mono[,2]),locations=rep(pop,nrow(d)))
    class(r) <-  c("simapop","GenomicHaplotypes")
    r
  }
  cat("Reading haplotypes from chromosome",chromosome,"for populations",paste(pops),"\n")

  ## first read in the legends
  if (!collapse) return(lapply(pops,read.phasing))
  else {
    l <- length(pops)
    if (l==1) {
      return(read.phasing(pops))
    }  else if (l==2) {
      return(combine.haplotypes(read.phasing(pops[1]),read.phasing(pops[2])))
    } else if (l==3) {
      return(combine.haplotypes(read.hapmap.phasing(chromosome,pops[1:2],collapse=TRUE)
                                ,read.phasing(pops[3])))
    } else {
      ha <- floor(l/2)
      return(combine.haplotypes(read.hapmap.phasing(chromosome,pops[1:ha],collapse=TRUE)
                                 ,read.hapmap.phasing(chromosome,pops[(ha+1):l],collapse=TRUE)))
    }
  }
}


##
print.hapmap.haplotypes <- function(h)
  {
    cat("HAPMAP phasing from populations",names(h$positions),"\n")
    for (i in 1:length(h$positions)) {
      cat(ncol(h$d[[i]]),"SNPs and ",nrow(h$d[[i]]),"haplotyes from population",names(h$positions)[i],"\n")
    }
  }
##

SNPmatch <- function(SNP1,SNP2)
  {
    a <- matrix(unlist(strsplit(SNP1,"/")),nrow=2)
    b <- matrix(unlist(strsplit(SNP2,"/")),nrow=2)
    return((a[1,]==b[1,]&a[2,]==b[2,])|
           (a[1,]==b[2,]&a[2,]==b[1,]))
    
  }
SNPmatchmono <- function(SNP,SNPmono)
  {
    a <- matrix(unlist(strsplit(SNP,"/")),nrow=2)
    return((a[1,]==SNPmono)|(a[2,]==SNPmono))
  }
### Add a population to the GenomicHaplotypes class
### that is currently in x
combine.haplotypes <- function(x1,x2)
  {
    nx1 <- nrow(x1$d)
    nx2 <- nrow(x2$d)
  
    ## get the various classes of SNPs to add
    ## both polymorphic
    bothpoly <- x1$positions %in% x2$positions
    u2 <- x2$positions %in% x1$positions  # reverse
    if (sum(x1$positions[bothpoly]!=x2$positions[u2])!=0) {
      print(paste("have found", sum(x1$positions[bothpoly]),"positions, but they do not match"))
      stop("problem in combine haplotypes")
    }
    dboth <- rbind(x1$d[,bothpoly],x2$d[,u2])
    posboth <- x1$positions[bothpoly]
    SNPboth <- x1$SNP[bothpoly]
    ## check that the SNPs are the same way round
    swp <- SNPboth!=x2$SNP[u2]
    dboth[,swp] <- 1-dboth[,swp]
    ## now see if they match!
    mt <- SNPmatch(SNPboth,x2$SNP[u2])
    if (sum(mt)!=length(SNPboth)) {
        wh <- which(mt==FALSE)
      warning(paste("mismatches in SNPs at positions",
                    paste(wh,collapse=" "),"values are",paste(SNPboth[wh],x2$SNP[u2][wh],sep="-"),"removing\n",collapse=" "))
      SNPboth <- SNPboth[mt]
      posboth <- posboth[mt]
      dboth <- dboth[,mt]
    }
    
    ## first mono, second poly
    monopoly <- x1$monopos %in% x2$positions
    u2 <- x2$positions %in%  x1$monopos
    SNPmonopoly <- x2$SNP[u2]
    val <- as.integer(substr(SNPmonopoly,3,3)==x1$monoSNP[monopoly])
    dtmp <- matrix(rep(val,nx1),byrow=T,nrow=nx1)
    dmonopoly <- rbind(dtmp,x2$d[,u2,drop=FALSE])
    posmonopoly <- x1$monopos[monopoly]
    mt <- SNPmatchmono(SNPmonopoly,x1$monoSNP[monopoly])
     
    if (sum(mt)!=length(SNPmonopoly)) {
      wh <- which(mt==FALSE)
      warning(paste("mismatches in SNPs at positions",
                   paste(wh,collapse=" "),"values are",paste(SNPmonopoly[wh],x1$monoSNP[monopoly],sep="-"),"removing\n",collapse=" "))
      dmonopoly <- dmonopoly[,mt]
      posmonopoly <- posmonopoly[mt]
      SNPmonopoly <- SNPmonopoly[mt]
    }
    
    ## first poly second mono
    polymono <- x1$positions %in% x2$monopos
    u2 <- x2$monopos %in%  x1$positions
    SNPpolymono <- x1$SNP[polymono]
    pospolymono <- x1$positions[polymono]
    val <- as.integer(substr(SNPpolymono,3,3)==x2$monoSNP[u2])
    dtmp <- matrix(rep(val,nx2),byrow=T,nrow=nx2)
    dpolymono <- rbind(x1$d[,polymono,drop=FALSE],dtmp)
      

    ## both mono 
    monomono <- x1$monopos %in% x2$monopos
    u2 <- x2$monopos %in% x1$monopos

    ## and different
    SNP1 <-  x1$monoSNP[monomono]
    SNP2 <- x2$monoSNP[u2]
    monomonodiff <- SNP1!=SNP2
    ndiff <- sum(monomonodiff)

    if (ndiff>0) {
      SNPmonomonodiff <- paste(SNP1[monomonodiff],SNP2[monomonodiff],sep="/")
      posmonomonodiff <- x1$monopos[monomono][monomonodiff]
      print(cbind(SNPmonomonodiff,posmonomonodiff))
      dmonomonodiff <- matrix(c(rep(0,ndiff*nx1),rep(1,ndiff*nx2)),ncol=ndiff,byrow=T)
    } else {
      SNPmonomonodiff <- NULL
      posmonomonodiff <- NULL
      dmonomonodiff <- NULL
    }

    ## fix the monos 
    monoSNP <- x1$monoSNP[!monomonodiff]
    monopos <- x1$monopos[!monomonodiff]

    ## and collect
    position <- c(posboth,posmonopoly,pospolymono,posmonomonodiff)
    SNP <- c(SNPboth,SNPmonopoly,SNPpolymono,SNPmonomonodiff)
    d <- cbind(dboth,dmonopoly,dpolymono,dmonomonodiff)
    
    o <- order(position)
#    print(paste(length(SNPboth),length(SNPmonopoly),length(SNPpolymono),length(SNPmonomonodiff)))
                                        #    print(paste(length(posboth),length(posmonopoly),length(pospolymono),length(posmonomonodiff)))
    
    r <- list(d=d[,o],SNP=SNP[o],positions=position[o],monoSNP=monoSNP
              ,monopos=monopos,location=factor(c(paste(x1$location),paste(x2$location))))
    class(r) <-  c("simapop","GenomicHaplotypes");
    r
  }


extract.hapmap.haplotypes <- function(h,start=0,finish=1,usepos)
  {
    if (missing(usepos)) {
      pos <- lapply(h$positions,function(x) x[x>start*10^6&x<=finish*10^6])
      usepos <- pos[[1]];
      if (length(pos)>1) {
        for (i in 2:length(pos)) {
          u <- match(usepos,pos[[i]])
          u <- u[!is.na(u)]
          usepos <- pos[[i]][u];
        }
        pos <- h$positions[[1]];
       # posa <- pos[pos>start*10^6&pos<=finish*10^6]   
      }
    }
      
    res <- NULL;
    loc <- NULL; 
    ## make sure that we have the same SNPs in all
    m <- match(usepos,h$positions[[1]])
    ## and make sure that the SNPs match use the first as a reference
    SNP=h$SNP[[1]][m]
    nSNP <- length(SNP)
    
    if (nSNP==0)
      stop(paste("no SNPs selected in the range [",start,",",finish,"]\n"))

    for (i in 1:length(h$d)) {
      m <- match(usepos,h$positions[[i]]);
      if (sum(is.na(m))>0) {
        stop(paste("missing positions",paste(usepos[is.na(m)],collapse=" "),"in population",names(h$positions)[i]));
      }
      thisres <- h$d[[i]][,m,drop=FALSE]
      thisSNP <- h$SNP[[i]][m]
      if (length(thisSNP)!=nSNP)
        stop("problem with number of SNPs")

      if (sum(SNPmatch(thisSNP,SNP))!=nSNP) {
        wh <- which(SNPmatch(thisSNP,SNP)==FALSE)
        warning(paste("mismatches in SNPs at positions",
                   paste(wh,collapse=" "),"values are",paste(cbind(SNP[wh],thisSNP[wh]),collapse=" ")))
        ## remove the offending positions
        #SNP <- SNP[-wh]
       # thisSNP <- thisSNP[-wh]
       # res <- res[,-wh]
       # thisres <- thisres[,-wh]
       # usepos <- usepos[-wh]
      }
    
      swp <- SNP!=thisSNP

     # thisres <- h$d[[i]][,m]
      thisres[,swp] <- 1-thisres[,swp]
      res <- rbind(res,thisres)
      loc <- c(loc,rep(names(h$positions)[i],nrow(h$d[[i]])))
    }
  
    if (length(h$d)==1) {
       res <- list(d=res,positions=usepos,SNP=SNP);
       class(res) <- "GenomicHaplotypes";
     } else {
       res <- list(d=res,positions=usepos,locations=factor(loc),SNP=SNP);
       class(res) <- c("simapop","GenomicHaplotypes");
     }
    res;
  }
## the following function only works for HCB and JPT data - the others need to look at the ped file
read.hapmap.genotypes <- function(chromosome,pop="HCB",positions=c(180248577,180253670,180256726,180257150))
{
  filename <- paste("/data/hapmap/genotypes_chr",chromosome,"_",pop,".txt.gz",sep="")
  dat <- read.table(gzfile(filename),skip=1,header=F)
  m <- match(positions,dat[,4])
  t(dat[m,12:56])
}




##
trycompare.genomic <- function(hm,pg)
  {
    u <- match(hm$positions,pg$positions);
  #  print(u)
    u1 <- which(!is.na(u));
    u2 <- u[!is.na(u)]

    which <- factor(c(rep("HM",nrow(hm$d)),rep("PG",nrow(pg$d))))

    ## which do we need to be careful with
    u <- hm$SNP[u1]!=pg$SNP[u2] 
    if (sum(u)>0) {
      ## separate out the alleles
      a <-  matrix(unlist(strsplit(hm$SNP[u1],"/")),ncol=2,byrow=T)
      b <- matrix(unlist(strsplit(pg$SNP[u2],"/")),ncol=2,byrow=T);
      ## first find the simple swaps
      swp1 <- (a[,1]==b[,2]&a[,2]==b[,1])
      ## now find those that are complementary  
      cmp1 <- (a[,1]=="A"&a[,2]=="G")&(b[,1]=="T"&b[,2]=="C")
      cmp2 <- (b[,1]=="A"&b[,2]=="G")&(a[,1]=="T"&a[,2]=="C")
      cmp3 <- (b[,1]=="G"&b[,2]=="A")&(a[,1]=="C"&a[,2]=="T")
      cmp4 <- (a[,1]=="G"&a[,2]=="A")&(b[,1]=="C"&a[,2]=="T")

    #  cmp5 <- (a[,1]=="A"&a[,2]=="C")&(b[,1]=="T"&b[,2]=="G")
      
      cmp <- cmp1|cmp2|cmp3|cmp4|cmp5
      ## now for those complementary and reversed
      cmp1 <- (a[,1]=="A"&a[,2]=="G")&(b[,1]=="C"&b[,2]=="T")
      cmp2 <- (b[,1]=="A"&b[,2]=="G")&(a[,1]=="C"&a[,2]=="T")
      cmp3 <- (b[,1]=="G"&b[,2]=="A")&(a[,1]=="T"&a[,2]=="C")
      cmp4 <- (a[,1]=="G"&a[,2]=="A")&(b[,1]=="T"&a[,2]=="C")
      cmprev <- cmp1|cmp2|cmp3|cmp4
      canuse <- (!u)|swp1|cmp|cmprev
      if (sum(canuse)!=length(u)) {
        cat("problems with alleles\n ");
        print(cbind(a[!canuse,],b[!canuse,]))
      } 
    
      d1 <- hm$d[,u1][,canuse]
      d2 <- pg$d[,u2][,canuse];
      d1[,swp1|cmprev] <- 1- d1[,swp1|cmprev]
    } else {
      d1 <- hm$d[,u1];
      d2 <- pg$d[,u2];
    }
    
    freq1 <- apply(d1,2,mean);
    freq2 <- apply(d2,2,mean);

    ## d2[,swp] <- 1-d2[,swp]
   #    print(freq1)
    
    print(cbind(hm$positions[u1],hm$SNP[u1],round(freq1,3)
                ,pg$positions[u2],pg$SNP[u2],round(freq2,3)))
    
    data <- c(apply(d1,1,paste,sep="",collapse=""),apply(d2,1,paste,sep="",collapse=""))
    locations <- factor(c(levels(hm$locations)[hm$locations],
                          levels(pg$locations)[pg$locations]))
    table(factor(data),locations,which)
  }


